import {Endpoint, EndpointOperations} from './endpoint';
import {Operation} from './operation';

export interface BaseAppState {
	endpoints: unknown;
	operations: unknown;
}

export type AppOperations<T extends App> = (
	T extends App<string, infer R>
		? R extends BaseAppState
			? R['operations'] extends []
				? R['operations'] extends readonly string[]
					? R['operations'][number]
					: string
				: string
			: never
		: never
	);

export interface App<AppName extends string = string, AppState extends BaseAppState = BaseAppState> {
	name: AppName;
	operations: Set<Operation>;
	endpoints: Set<Endpoint>;
	operation<NewOperation extends Operation>(newOperation: NewOperation): App<
		AppName,
		{
			endpoints: AppState['endpoints'],
			operations: AppState['operations'] extends Array<unknown> ? [...AppState['operations'], NewOperation['name']] : [NewOperation['name']]
		}
	>;
	endpoint<NewEndpoint extends Endpoint = Endpoint>(
		newEndpoint: EndpointOperations<NewEndpoint> extends AppOperations<this>
			? NewEndpoint
			: never
	): App<
		AppName,
		{
			endpoints: AppState['endpoints'] extends Array<unknown> ? [...AppState['endpoints'], NewEndpoint] : [NewEndpoint],
			operations: AppState['operations']
		}
	>;
}

interface AppParams<Name extends string = string> {
	name: Name;
}

class AppInstance<Params extends AppParams, State extends BaseAppState> implements App<Params['name'], State> {
	readonly name: Params['name'];
	readonly endpoints: Set<Endpoint>;
	readonly operations: Set<Operation>;

	constructor(params: Params) {
		this.name = params.name;
		this.endpoints = new Set<Endpoint>();
		this.operations = new Set<Operation>();
	}

	operation<NewOperation extends Operation>(newOperation: NewOperation) {
		this.operations.add(newOperation);

		return this;
	}

	endpoint<NewEndpoint extends Endpoint = Endpoint>(newEndpoint: NewEndpoint) {
		this.endpoints.add(newEndpoint);

		return this;
	}
}

export const app = <Params extends AppParams>(params: Params): App<Params['name'], {
	endpoints: [];
	operations: [];
}> => {
	return new AppInstance(params);
};
