export const AVAILABLE_METHODS = [
	'HEAD',
	'GET',
	'POST',
	'PUT',
	'PATCH',
	'DELETE',
	'OPTIONS'
] as const;

export const AVAILABLE_EXTENSION_METHODS = [
	'QUERY'
] as const;

export type Method = typeof AVAILABLE_METHODS[number];

export type MethodWithExtensions = Method | typeof AVAILABLE_EXTENSION_METHODS[number];

export interface BaseOperationParams<
	Name extends string = string,
	Method extends MethodWithExtensions = MethodWithExtensions,
> {
	name: Name;
	method?: Method;
}

export interface Operation<Params extends BaseOperationParams = BaseOperationParams> {
	name: Params['name'];
	method: Params['method'];
}

class OperationInstance<Params extends BaseOperationParams = BaseOperationParams> implements Operation<Params> {
	readonly name: Params['name'];
	readonly method: Params['method'];

	constructor(params: Params) {
		this.name = params.name;
		this.method = params.method ?? 'GET';
	}
}

export const operation = <Params extends BaseOperationParams = BaseOperationParams>(
	params: Params
): Operation<Params> => {
	return new OperationInstance(params);
};
